/* Copyright Airship and Contributors */

#import "UAComponent.h"
#import "UAChannelNotificationCenterEvents.h"
#import "UAAttributeMutations.h"
#import "UATagGroupsMutation.h"
#import "UAAttributePendingMutations.h"

NS_ASSUME_NONNULL_BEGIN

//---------------------------------------------------------------------------------------
// UAChannel Class
//---------------------------------------------------------------------------------------
/**
 * Notification posted when a channel tag group mutation is uploaded.
 *
 * User data will contain a UATagGroupsMutation, identifier string and NSDate under
 * UAChannelUploadedAudienceMutationNotificationMutationKey,
 * UAChannelUploadedAudienceMutationNotificationIdentifierKey, and
 * UAChannelUploadedAudienceMutationNotificationDateKey, respectively.
 * @note For internal use only. :nodoc:
 */
extern NSString *const UAChannelUploadedTagGroupMutationNotification;

/**
 * Notification posted when a channel attribute mutation is uploaded.
 *
 * User data will contain a UAAttributePendingMutations, identifier string and NSDate under
 * UAChannelUploadedAudienceMutationNotificationMutationKey,
 * UAChannelUploadedAudienceMutationNotificationIdentifierKey, and
 * UAChannelUploadedAudienceMutationNotificationDateKey, respectively.
 * @note For internal use only. :nodoc:
 */
extern NSString *const UAChannelUploadedAttributeMutationsNotification;

/**
 * The mutation key for UAChannelUploadedTagGroupMutationNotification and UAChannelUploadedAttributeMutationsNotification.
 * @note For internal use only. :nodoc:
 */
extern NSString *const UAChannelUploadedAudienceMutationNotificationMutationKey;

/**
 * The identifier key for UAChannelUploadedTagGroupMutationNotification and UAChannelUploadedAttributeMutationsNotification.
 * @note For internal use only. :nodoc:
 */
extern NSString *const UAChannelUploadedAudienceMutationNotificationIdentifierKey;

/**
 * The date key for UAChannelUploadedTagGroupMutationNotification and UAChannelUploadedAttributeMutationsNotification.
 * @note For internal use only. :nodoc:
 */
extern NSString *const UAChannelUploadedAudienceMutationNotificationDateKey;

/**
* This singleton provides an interface to the channel functionality provided by the Airship iOS Push API.
*/
@interface UAChannel : UAComponent

/**
 The Channel ID.
 */
@property(nullable, nonatomic, readonly) NSString *identifier;

/**
 * Returns the pending tag groups mutuations.
 * @note For internal use only. :nodoc:
 */
@property (nonatomic, readonly)NSArray<UATagGroupsMutation *> *pendingTagGroups;

/**
 * Returns the pending attribute mutuations.
 * @note For internal use only. :nodoc:
 */
@property (nonatomic, readonly) UAAttributePendingMutations *pendingAttributes;

/**
 Tags for this device.
 */
@property (nonatomic, copy) NSArray<NSString *> *tags;

/**
 * Allows setting tags from the device. Tags can be set from either the server or the device, but
 * not both (without synchronizing the data), so use this flag to explicitly enable or disable
 * the device-side flags.
 *
 * Set this to `NO` to prevent the device from sending any tag information to the server when using
 * server-side tagging. Defaults to `YES`.
 */
@property (nonatomic, assign, getter=isChannelTagRegistrationEnabled) BOOL channelTagRegistrationEnabled;

///---------------------------------------------------------------------------------------
/// @name Tags
///---------------------------------------------------------------------------------------

/**
 * Adds a tag to the list of tags for the device. To update the server, make all of your changes, then call
 * `updateRegistration` to update the Airship server.
 *
 * @note When updating multiple server-side values (tags, alias, time zone, quiet time), set the
 * values first, then call `updateRegistration`. Batching these calls improves performance.
 *
 * @param tag Tag to be added
 */
- (void)addTag:(NSString *)tag;

/**
 * Adds a collection of tags to the current list of device tags. To update the server, make all of your
 * changes, then call `updateRegistration`.
 *
 * @note When updating multiple server-side values (tags, alias, time zone, quiet time), set the
 * values first, then call `updateRegistration`. Batching these calls improves performance.
 *
 * @param tags Array of new tags
 */
- (void)addTags:(NSArray<NSString *> *)tags;

/**
 * Removes a tag from the current tag list. To update the server, make all of your changes, then call
 * `updateRegistration`.
 *
 * @note When updating multiple server-side values (tags, alias, time zone, quiet time), set the
 * values first, then call `updateRegistration`. Batching these calls improves performance.
 *
 * @param tag Tag to be removed
 */
- (void)removeTag:(NSString *)tag;

/**
 * Removes a collection of tags from a device. To update the server, make all of your changes, then call
 * `updateRegistration`.
 *
 * @note When updating multiple server-side values (tags, alias, time zone, quiet time), set the
 * values first, then call `updateRegistration`. Batching these calls improves performance.
 *
 * @param tags Array of tags to be removed
 */
- (void)removeTags:(NSArray<NSString *> *)tags;


///---------------------------------------------------------------------------------------
/// @name Tag Groups
///---------------------------------------------------------------------------------------

/**
 * Add tags to channel tag groups. To update the server,
 * make all of your changes, then call `updateRegistration`.
 *
 * @param tags Array of tags to add.
 * @param tagGroupID Tag group ID string.
 */
- (void)addTags:(NSArray<NSString *> *)tags group:(NSString *)tagGroupID;


/**
 * Removes tags from channel tag groups. To update the server,
 * make all of your changes, then call `updateRegistration`.
 *
 * @param tags Array of tags to remove.
 * @param tagGroupID Tag group ID string.
 */
- (void)removeTags:(NSArray<NSString *> *)tags group:(NSString *)tagGroupID;

/**
 * Sets tags for channel tag groups. To update the server,
 * make all of your changes, then call `updateRegistration`.
 *
 * @param tags Array of tags to set.
 * @param tagGroupID Tag group ID string.
 */
- (void)setTags:(NSArray<NSString *> *)tags group:(NSString *)tagGroupID;


///---------------------------------------------------------------------------------------
/// @name Channel Attributes
///---------------------------------------------------------------------------------------

/**
 * Applies mutations to attributes associated with this device.
 *
 * @param mutations Attribute mutations to apply to this device.
 */
- (void)applyAttributeMutations:(UAAttributeMutations *)mutations;

///---------------------------------------------------------------------------------------
/// @name Channel Registration
///---------------------------------------------------------------------------------------

/**
 * Enables channel creation if channelCreationDelayEnabled was set to `YES` in the config.
 */
- (void)enableChannelCreation;

/**
 * Registers or updates the current registration with an API call. If push notifications are
 * not enabled, this unregisters the device token.
 *
 * Observe NSNotificationCenterEvents such as UAChannelCreatedEvent, UAChannelUpdatedEvent and UAChannelRegistrationFailedEvent
 * to receive success and failure callbacks.
 */
- (void)updateRegistration;

@end

NS_ASSUME_NONNULL_END
